# Copyright (c) 2021 Emanuele Bellocchia
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""Module with helper class for Litecoin configuration handling."""

# Imports
from typing import Any, Dict, Type

from bip_utils.addr import IAddrEncoder
from bip_utils.bip.bip32 import Bip32KeyNetVersions
from .bip_coin_conf import Bip32Base, BipCoinConf
from bip_utils.utils.conf import CoinNames


class BipLitecoinConf(BipCoinConf):
    """
    Litecoin configuration class.
    It allows to return different addresses and key net versions depending on the configuration.
    """

    m_alt_key_net_ver: Bip32KeyNetVersions
    m_use_alt_key_net_ver: bool
    m_use_depr_addr: bool

    def __init__(self,  # pylint: disable=too-many-arguments
                 coin_names: CoinNames,
                 coin_idx: int,
                 is_testnet: bool,
                 def_path: str,
                 key_net_ver: Bip32KeyNetVersions,
                 alt_key_net_ver: Bip32KeyNetVersions,
                 wif_net_ver: bytes,
                 bip32_cls: Type[Bip32Base],
                 addr_cls: Type[IAddrEncoder],
                 addr_params: Dict[str, Any]) -> None:
        """
        Construct class.

        Args:
            coin_names (CoinNames object)               : Coin names
            coin_idx (int)                              : Coin index
            is_testnet (bool)                           : Test net flag
            def_path (str)                              : Default path
            key_net_ver (Bip32KeyNetVersions object)    : Key net versions
            alt_key_net_ver (Bip32KeyNetVersions object): Key net versions (alternate)
            wif_net_ver (bytes)                         : WIF net version
            bip32_cls (Bip32Base class)                 : Bip32 class
            addr_params (dict)                          : Address parameters
            addr_cls (IAddrEncoder class)               : Address class
        """
        super().__init__(coin_names=coin_names,
                         coin_idx=coin_idx,
                         is_testnet=is_testnet,
                         def_path=def_path,
                         key_net_ver=key_net_ver,
                         wif_net_ver=wif_net_ver,
                         bip32_cls=bip32_cls,
                         addr_cls=addr_cls,
                         addr_params=addr_params)

        self.m_alt_key_net_ver = alt_key_net_ver
        self.m_use_alt_key_net_ver = False
        self.m_use_depr_addr = False

    def UseAlternateKeyNetVersions(self,
                                   value: bool) -> None:
        """
        Select if use the alternate key net version.

        Args:
            value (bool): True for using alternate key net version, false for using the standard one
        """
        self.m_use_alt_key_net_ver = value

    def UseDeprecatedAddress(self,
                             value: bool) -> None:
        """
        Select if use the deprecated address.

        Args:
            value (bool): True for using deprecated address, false for using the standard one
        """
        self.m_use_depr_addr = value

    def KeyNetVersions(self) -> Bip32KeyNetVersions:
        """
        Get key net versions. It overrides the method in BipCoinConf.
        Litecoin overrides the method because it can have 2 different key net versions.

        Returns:
            Bip32KeyNetVersions object: Bip32KeyNetVersions object
        """
        return self.m_alt_key_net_ver if self.m_use_alt_key_net_ver else self.m_key_net_ver

    def AddrParams(self) -> Dict[str, Any]:
        """
        Get the address parameters. It overrides the method in BipCoinConf.

        Returns:
            dict: Address parameters
        """
        return ({"net_ver": self.m_addr_params["depr_net_ver"]}
                if self.m_use_depr_addr
                else {"net_ver": self.m_addr_params["std_net_ver"]})



# Copyright (c) 2021 Emanuele Bellocchia
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""Module for BIP44 coins enum."""

# Imports
from enum import auto, unique

from ..common.bip_coins import BipCoins


@unique
class Bip44Coins(BipCoins):
    """Enumerative for supported BIP44 coins."""

    # Main nets
    AKASH_NETWORK = auto()
    ALGORAND = auto()
    APTOS = auto()
    ARBITRUM = auto()
    AVAX_C_CHAIN = auto()
    AVAX_P_CHAIN = auto()
    AVAX_X_CHAIN = auto()
    AXELAR = auto()
    BAND_PROTOCOL = auto()
    BINANCE_CHAIN = auto()
    BINANCE_SMART_CHAIN = auto()
    BITCOIN = auto()
    BITCOIN_CASH = auto()
    BITCOIN_CASH_SLP = auto()
    BITCOIN_SV = auto()
    CARDANO_BYRON_ICARUS = auto()
    CARDANO_BYRON_LEDGER = auto()
    CELESTIA = auto()
    CELO = auto()
    CERTIK = auto()
    CHIHUAHUA = auto()
    COSMOS = auto()
    DASH = auto()
    DOGECOIN = auto()
    DYDX = auto()
    ECASH = auto()
    ELROND = auto()
    EOS = auto()
    ERGO = auto()
    ETHEREUM = auto()
    ETHEREUM_CLASSIC = auto()
    FANTOM_OPERA = auto()
    FETCH_AI = auto()
    FETCH_AI_ETH = auto()
    FILECOIN = auto()
    HARMONY_ONE_ATOM = auto()
    HARMONY_ONE_ETH = auto()
    HARMONY_ONE_METAMASK = auto()
    HUOBI_CHAIN = auto()
    ICON = auto()
    INJECTIVE = auto()
    IRIS_NET = auto()
    KAVA = auto()
    KUSAMA_ED25519_SLIP = auto()
    LITECOIN = auto()
    METIS = auto()
    MONERO_ED25519_SLIP = auto()
    MONERO_SECP256K1 = auto()
    MULTIVERSX = auto()
    NANO = auto()
    NEAR_PROTOCOL = auto()
    NEO = auto()
    NEO_LEGACY = auto()
    NEO_N3 = auto()
    NEUTRON = auto()
    NIMIQ = auto()
    NINE_CHRONICLES_GOLD = auto()
    OKEX_CHAIN_ATOM = auto()
    OKEX_CHAIN_ATOM_OLD = auto()
    OKEX_CHAIN_ETH = auto()
    ONTOLOGY = auto()
    OPTIMISM = auto()
    OSMOSIS = auto()
    PI_NETWORK = auto()
    POLKADOT_ED25519_SLIP = auto()
    POLYGON = auto()
    RIPPLE = auto()
    SECRET_NETWORK_OLD = auto()
    SECRET_NETWORK_NEW = auto()
    SOLANA = auto()
    STAFI = auto()
    STELLAR = auto()
    SUI = auto()
    TERRA = auto()
    TEZOS = auto()
    THETA = auto()
    TRON = auto()
    VECHAIN = auto()
    VERGE = auto()
    ZCASH = auto()
    ZILLIQA = auto()
    # Test nets
    BITCOIN_CASH_TESTNET = auto()
    BITCOIN_CASH_SLP_TESTNET = auto()
    BITCOIN_SV_TESTNET = auto()
    BITCOIN_REGTEST = auto()
    BITCOIN_TESTNET = auto()
    DASH_TESTNET = auto()
    DOGECOIN_TESTNET = auto()
    ECASH_TESTNET = auto()
    ERGO_TESTNET = auto()
    LITECOIN_TESTNET = auto()
    ZCASH_TESTNET = auto()



# Common
from .common.ikeys import IPrivateKey, IPublicKey
from .common.ipoint import IPoint

# Curve
from .curve.elliptic_curve import EllipticCurve
from .curve.elliptic_curve_getter import EllipticCurveGetter
from .curve.elliptic_curve_types import EllipticCurveTypes

# ed25519
# from .ed25519.ed25519 import Ed25519
# from .ed25519.ed25519_keys import Ed25519PrivateKey, Ed25519PublicKey
# from .ed25519.ed25519_point import Ed25519Point
# from .ed25519.ed25519_utils import Ed25519Utils

# ed25519-blake2b
# from .ed25519_blake2b.ed25519_blake2b import Ed25519Blake2b
# from .ed25519_blake2b.ed25519_blake2b_keys import Ed25519Blake2bPrivateKey, Ed25519Blake2bPublicKey
# from .ed25519_blake2b.ed25519_blake2b_point import Ed25519Blake2bPoint

# ed25519-kholaw
# from .ed25519_kholaw.ed25519_kholaw import Ed25519Kholaw
# from .ed25519_kholaw.ed25519_kholaw_keys import Ed25519KholawPrivateKey, Ed25519KholawPublicKey
# from .ed25519_kholaw.ed25519_kholaw_point import Ed25519KholawPoint

# ed25519-monero
# from .ed25519_monero.ed25519_monero import Ed25519Monero
# from .ed25519_monero.ed25519_monero_keys import Ed25519MoneroPrivateKey, Ed25519MoneroPublicKey
# from .ed25519_monero.ed25519_monero_point import Ed25519MoneroPoint

# nist256p1
# from .nist256p1.nist256p1 import Nist256p1
# from .nist256p1.nist256p1_keys import Nist256p1PrivateKey, Nist256p1PublicKey
# from .nist256p1.nist256p1_point import Nist256p1Point

# secp256k1
from .secp256k1.secp256k1 import Secp256k1, Secp256k1Point, Secp256k1PrivateKey, Secp256k1PublicKey

# sr25519
# from .sr25519.sr25519 import Sr25519
# from .sr25519.sr25519_keys import Sr25519PrivateKey, Sr25519PublicKey
# from .sr25519.sr25519_point import Sr25519Point



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: dydxprotocol/clob/order_removals.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from ...gogoproto import gogo_pb2 as gogoproto_dot_gogo__pb2
from ..clob import order_pb2 as dydxprotocol_dot_clob_dot_order__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n&dydxprotocol/clob/order_removals.proto\x12\x11\x64ydxprotocol.clob\x1a\x14gogoproto/gogo.proto\x1a\x1d\x64ydxprotocol/clob/order.proto\"\xe8\x04\n\x0cOrderRemoval\x12\x32\n\x08order_id\x18\x01 \x01(\x0b\x32\x1a.dydxprotocol.clob.OrderIdB\x04\xc8\xde\x1f\x00\x12\x45\n\x0eremoval_reason\x18\x02 \x01(\x0e\x32-.dydxprotocol.clob.OrderRemoval.RemovalReason\"\xdc\x03\n\rRemovalReason\x12\x1e\n\x1aREMOVAL_REASON_UNSPECIFIED\x10\x00\x12&\n\"REMOVAL_REASON_UNDERCOLLATERALIZED\x10\x01\x12&\n\"REMOVAL_REASON_INVALID_REDUCE_ONLY\x10\x02\x12\x34\n0REMOVAL_REASON_POST_ONLY_WOULD_CROSS_MAKER_ORDER\x10\x03\x12%\n!REMOVAL_REASON_INVALID_SELF_TRADE\x10\x04\x12<\n8REMOVAL_REASON_CONDITIONAL_FOK_COULD_NOT_BE_FULLY_FILLED\x10\x05\x12\x35\n1REMOVAL_REASON_CONDITIONAL_IOC_WOULD_REST_ON_BOOK\x10\x06\x12\x1f\n\x1bREMOVAL_REASON_FULLY_FILLED\x10\x07\x12;\n7REMOVAL_REASON_VIOLATES_ISOLATED_SUBACCOUNT_CONSTRAINTS\x10\x08\x12+\n\'REMOVAL_REASON_PERMISSIONED_KEY_EXPIRED\x10\tB8Z6github.com/dydxprotocol/v4-chain/protocol/x/clob/typesb\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'dydxprotocol.clob.order_removals_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'Z6github.com/dydxprotocol/v4-chain/protocol/x/clob/types'
  _ORDERREMOVAL.fields_by_name['order_id']._options = None
  _ORDERREMOVAL.fields_by_name['order_id']._serialized_options = b'\310\336\037\000'
  _ORDERREMOVAL._serialized_start=115
  _ORDERREMOVAL._serialized_end=731
  _ORDERREMOVAL_REMOVALREASON._serialized_start=255
  _ORDERREMOVAL_REMOVALREASON._serialized_end=731
# @@protoc_insertion_point(module_scope)



# Copyright (c) 2022 Emanuele Bellocchia
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""Module for Scrypt algorithm."""

# Imports
from typing import Union

from Crypto.Protocol.KDF import scrypt

from ..misc import AlgoUtils


class Scrypt:
    """
    Scrypt class.
    It derives key using Scrypt algorithm.
    """

    @staticmethod
    def DeriveKey(password: Union[bytes, str],  # pylint: disable=too-many-arguments
                  salt: Union[bytes, str],
                  key_len: int,
                  n: int,
                  r: int,
                  p: int) -> bytes:
        """
        Derive a key.

        Args:
            password (str or bytes): Password
            salt (str or bytes)    : Salt
            key_len (int)          : Length of the derived key
            n (int)                : CPU/Memory cost parameter
            r (int)                : Block size parameter
            p (int)                : Parallelization parameter

        Returns:
            bytes: Computed result
        """

        # Type for password and salt should be Union[bytes, str] in pycryptodome, but it's only str
        # So, we ignore the mypy warning
        return scrypt(AlgoUtils.Encode(password),   # type: ignore [arg-type, return-value]
                      AlgoUtils.Encode(salt),       # type: ignore [arg-type]
                      key_len=key_len,
                      N=n,
                      r=r,
                      p=p)



# Copyright (c) 2021 Emanuele Bellocchia
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""
Module for BIP44 keys derivation.
Reference: https://github.com/bitcoin/bips/blob/master/bip-0044.mediawiki
"""

# Imports
from typing import Union

from ..bip32 import Bip32KeyData, Bip32KeyIndex, Bip32KeyNetVersions, Bip32Const, Bip32Slip10Secp256k1
from ..bip44_base import Bip44Base, Bip44Changes, Bip44Levels

_BIP44_BTC_KEY_NET_VER_MAIN: Bip32KeyNetVersions = Bip32Const.MAIN_NET_KEY_NET_VERSIONS

# from ..conf.bip44 import Bip44ConfGetter
# from ..conf.bip44 import BipCoinConf
from ..conf.common import BipCoins, BipCoinConf
from ..coin_conf import CoinsConf
from ..slip.slip44 import Slip44
from ..conf.common import DER_PATH_NON_HARDENED_FULL
# from bip_utils.ecc import IPrivateKey, IPublicKey


class Bip44Const:
    """Class container for BIP44 constants."""

    # Specification name
    SPEC_NAME: str = "BIP-0044"
    # Purpose
    PURPOSE: int = Bip32KeyIndex.HardenIndex(44)


class Bip44(Bip44Base):
    """
    BIP44 class.
    It allows master key generation and children keys derivation in according to BIP-0044.
    """

    #
    # Class methods for construction
    #

    @classmethod
    def FromSeed(cls,
                 seed_bytes: bytes) -> Bip44Base:
        """
        Create a Bip44Base object from the specified seed (e.g. BIP39 seed).

        Args:
            seed_bytes (bytes)  : Seed bytes
            coin_type (BipCoins): Coin type, shall be a Bip44Coins enum

        Returns:
            Bip44Base object: Bip44Base object

        Raises:
            TypeError: If coin type is not a Bip44Coins enum
            ValueError: If the seed is too short
            Bip32KeyError: If the seed is not suitable for master key generation
        """

        # Bip44ConfGetter already checks the enum type
        conf = BipCoinConf(
            coin_names=CoinsConf.Cosmos.CoinNames(),
            coin_idx=Slip44.ATOM,
            is_testnet=False,
            def_path=DER_PATH_NON_HARDENED_FULL,
            key_net_ver=_BIP44_BTC_KEY_NET_VER_MAIN,
            wif_net_ver=None,
            bip32_cls=Bip32Slip10Secp256k1,
            # addr_cls=AtomAddrEncoder,
            addr_params={
                "hrp": CoinsConf.Cosmos.ParamByKey("addr_hrp"),
            },
        )
        return cls._FromSeed(seed_bytes,
                             conf)

    @classmethod
    def FromExtendedKey(cls,
                        ex_key_str: str) -> Bip44Base:
        """
        Create a Bip44Base object from the specified extended key.

        Args:
            ex_key_str (str)    : Extended key string
            coin_type (BipCoins): Coin type, shall be a Bip44Coins enum

        Returns:
            Bip44Base object: Bip44Base object

        Raises:
            TypeError: If coin type is not a Bip44Coins enum
            Bip32KeyError: If the extended key is not valid
        """

        # Bip44ConfGetter already checks the enum type
        return cls._FromExtendedKey(ex_key_str, Bip44ConfGetter.GetConfig(coin_type))

    # @classmethod
    # def FromPrivateKey(cls,
    #                    priv_key: Union[bytes, IPrivateKey],
    #                    coin_type: BipCoins,
    #                    key_data: Bip32KeyData = Bip32KeyData()) -> Bip44Base:
    #     """
    #     Create a Bip44Base object from the specified private key and derivation data.
    #     If only the private key bytes are specified, the key will be considered a master key with
    #     the chain code set to zero, since there is no way to recover the key derivation data.

    #     Args:
    #         priv_key (bytes or IPrivateKey)         : Private key
    #         coin_type (BipCoins)                    : Coin type, shall be a Bip44Coins enum
    #         key_data (Bip32KeyData object, optional): Key data (default: all zeros)

    #     Returns:
    #         Bip44Base object: Bip44Base object

    #     Raises:
    #         TypeError: If coin type is not a Bip44Coins enum
    #         Bip32KeyError: If the key is not valid
    #     """

    #     # Bip44ConfGetter already checks the enum type
    #     return cls._FromPrivateKey(priv_key,
    #                                Bip44ConfGetter.GetConfig(coin_type),
    #                                key_data)

    # @classmethod
    # def FromPublicKey(cls,
    #                   pub_key: Union[bytes, IPublicKey],
    #                   coin_type: BipCoins,
    #                   key_data: Bip32KeyData = Bip32KeyData(depth=Bip44Levels.ACCOUNT)) -> Bip44Base:
    #     """
    #     Create a Bip44Base object from the specified public key and derivation data.
    #     If only the public key bytes are specified, the key will be considered an account key with
    #     the chain code set to zero, since there is no way to recover the key derivation data.

    #     Args:
    #         pub_key (bytes or IPublicKey)           : Public key
    #         coin_type (BipCoins)                    : Coin type, shall be a Bip44Coins enum
    #         key_data (Bip32KeyData object, optional): Key data (default: all zeros with account depth)

    #     Returns:
    #         Bip44Base object: Bip44Base object

    #     Raises:
    #         TypeError: If coin type is not a Bip44Coins enum
    #         Bip32KeyError: If the key is not valid
    #     """

    #     # Bip44ConfGetter already checks the enum type
    #     return cls._FromPublicKey(pub_key,
    #                               Bip44ConfGetter.GetConfig(coin_type),
    #                               key_data)

    #
    # Overridden abstract methods
    #

    def Purpose(self) -> Bip44Base:
        """
        Derive a child key from the purpose and return a new Bip44Base object.

        Returns:
            Bip44Base object: Bip44Base object

        Raises:
            Bip44DepthError: If current depth is not suitable for deriving keys
            Bip32KeyError: If the derivation results in an invalid key
        """
        return self._PurposeGeneric(Bip44Const.PURPOSE)

    def Coin(self) -> Bip44Base:
        """
        Derive a child key from the coin type specified at construction and return a new Bip44Base object.

        Returns:
            Bip44Base object: Bip44Base object

        Raises:
            Bip44DepthError: If current depth is not suitable for deriving keys
            Bip32KeyError: If the derivation results in an invalid key
        """
        return self._CoinGeneric()

    def Account(self,
                acc_idx: int) -> Bip44Base:
        """
        Derive a child key from the specified account index and return a new Bip44Base object.

        Args:
            acc_idx (int): Account index

        Returns:
            Bip44Base object: Bip44Base object

        Raises:
            Bip44DepthError: If current depth is not suitable for deriving keys
            Bip32KeyError: If the derivation results in an invalid key
        """
        return self._AccountGeneric(acc_idx)

    def Change(self,
               change_type: Bip44Changes) -> Bip44Base:
        """
        Derive a child key from the specified change type and return a new Bip44Base object.

        Args:
            change_type (Bip44Changes): Change type, must a Bip44Changes enum

        Returns:
            Bip44Base object: Bip44Base object

        Raises:
            TypeError: If change type is not a Bip44Changes enum
            Bip44DepthError: If current depth is not suitable for deriving keys
            Bip32KeyError: If the derivation results in an invalid key
        """
        return self._ChangeGeneric(change_type)

    def AddressIndex(self,
                     addr_idx: int) -> Bip44Base:
        """
        Derive a child key from the specified address index and return a new Bip44Base object.

        Args:
            addr_idx (int): Address index

        Returns:
            Bip44Base object: Bip44Base object

        Raises:
            Bip44DepthError: If current depth is not suitable for deriving keys
            Bip32KeyError: If the derivation results in an invalid key
        """
        return self._AddressIndexGeneric(addr_idx)

    @staticmethod
    def SpecName() -> str:
        """
        Get specification name.

        Returns:
            str: Specification name
        """
        return Bip44Const.SPEC_NAME



from google.protobuf.json_format import ParseDict
from google.protobuf.any_pb2 import Any
from .dydxprotocol.sending.transfer_pb2 import (
    MsgDepositToSubaccount,
    MsgWithdrawFromSubaccount,
)
from .dydxprotocol.sending.tx_pb2 import MsgCreateTransfer
from .dydxprotocol.clob.tx_pb2 import (
    MsgPlaceOrder,
    MsgCancelOrder,
    MsgBatchCancel,
)
from .dydxprotocol.accountplus.tx_pb2 import TxExtension
from .cosmos.crypto.secp256k1.keys_pb2 import PubKey

registry = {
  '/dydxprotocol.clob.MsgPlaceOrder': MsgPlaceOrder,
  '/dydxprotocol.clob.MsgCancelOrder': MsgCancelOrder,
  '/dydxprotocol.clob.MsgBatchCancel': MsgBatchCancel,


  '/dydxprotocol.sending.MsgCreateTransfer': MsgCreateTransfer,
  '/dydxprotocol.sending.MsgWithdrawFromSubaccount': MsgWithdrawFromSubaccount,
  '/dydxprotocol.sending.MsgDepositToSubaccount': MsgDepositToSubaccount,
  '/dydxprotocol.accountplus.TxExtension': TxExtension,
  '/cosmos.crypto.secp256k1.PubKey': PubKey,
}

def encode_as_any(encodeObject):
    typeUrl = encodeObject['typeUrl']
    value = encodeObject['value']
    t = registry[typeUrl]
    message = ParseDict(value, t())
    packed = Any(
        type_url=typeUrl,
        value=message.SerializeToString()
    )
    return packed



import binascii

from . import ecdsa
from . import der
from . import rfc6979
from .curves import NIST192p, find_curve
from .ecdsa import RSZeroError
from .util import string_to_number, number_to_string, randrange
from .util import sigencode_string, sigdecode_string
from .util import oid_ecPublicKey, encoded_oid_ecPublicKey
from hashlib import sha1


class BadSignatureError(Exception):
    pass


class BadDigestError(Exception):
    pass


class VerifyingKey:
    def __init__(self, _error__please_use_generate=None):
        if not _error__please_use_generate:
            raise TypeError("Please use VerifyingKey.generate() to "
                            "construct me")

    @classmethod
    def from_public_point(klass, point, curve=NIST192p, hashfunc=sha1):
        self = klass(_error__please_use_generate=True)
        self.curve = curve
        self.default_hashfunc = hashfunc
        self.pubkey = ecdsa.Public_key(curve.generator, point)
        self.pubkey.order = curve.order
        return self

    @classmethod
    def from_string(klass, string, curve=NIST192p, hashfunc=sha1,
                    validate_point=True):
        order = curve.order
        assert (len(string) == curve.verifying_key_length), \
               (len(string), curve.verifying_key_length)
        xs = string[:curve.baselen]
        ys = string[curve.baselen:]
        assert len(xs) == curve.baselen, (len(xs), curve.baselen)
        assert len(ys) == curve.baselen, (len(ys), curve.baselen)
        x = string_to_number(xs)
        y = string_to_number(ys)
        if validate_point:
            assert ecdsa.point_is_valid(curve.generator, x, y)
        from . import ellipticcurve
        point = ellipticcurve.Point(curve.curve, x, y, order)
        return klass.from_public_point(point, curve, hashfunc)

    @classmethod
    def from_pem(klass, string):
        return klass.from_der(der.unpem(string))

    @classmethod
    def from_der(klass, string):
        # [[oid_ecPublicKey,oid_curve], point_str_bitstring]
        s1, empty = der.remove_sequence(string)
        if empty != b'':
            raise der.UnexpectedDER("trailing junk after DER pubkey: %s" %
                                    binascii.hexlify(empty))
        s2, point_str_bitstring = der.remove_sequence(s1)
        # s2 = oid_ecPublicKey,oid_curve
        oid_pk, rest = der.remove_object(s2)
        oid_curve, empty = der.remove_object(rest)
        if empty != b'':
            raise der.UnexpectedDER("trailing junk after DER pubkey objects: %s" %
                                    binascii.hexlify(empty))
        assert oid_pk == oid_ecPublicKey, (oid_pk, oid_ecPublicKey)
        curve = find_curve(oid_curve)
        point_str, empty = der.remove_bitstring(point_str_bitstring)
        if empty != b'':
            raise der.UnexpectedDER("trailing junk after pubkey pointstring: %s" %
                                    binascii.hexlify(empty))
        assert point_str.startswith(b'\x00\x04')
        return klass.from_string(point_str[2:], curve)

    @classmethod
    def from_public_key_recovery(klass, signature, data, curve, hashfunc=sha1, sigdecode=sigdecode_string):
        # Given a signature and corresponding message this function
        # returns a list of verifying keys for this signature and message

        digest = hashfunc(data).digest()
        return klass.from_public_key_recovery_with_digest(signature, digest, curve, hashfunc=sha1, sigdecode=sigdecode)

    @classmethod
    def from_public_key_recovery_with_digest(klass, signature, digest, curve, hashfunc=sha1, sigdecode=sigdecode_string):
        # Given a signature and corresponding digest this function
        # returns a list of verifying keys for this signature and message

        generator = curve.generator
        r, s = sigdecode(signature, generator.order())
        sig = ecdsa.Signature(r, s)

        digest_as_number = string_to_number(digest)
        pks = sig.recover_public_keys(digest_as_number, generator)

        # Transforms the ecdsa.Public_key object into a VerifyingKey
        verifying_keys = [klass.from_public_point(pk.point, curve, hashfunc) for pk in pks]
        return verifying_keys

    def _raw_encode(self):
        """Convert the point to the :term:`raw encoding`."""
        order = self.pubkey.order
        x_str = number_to_string(self.pubkey.point.x(), order)
        y_str = number_to_string(self.pubkey.point.y(), order)
        return x_str + y_str

    def _compressed_encode(self):
        """Encode the point into the compressed form."""
        order = self.pubkey.order
        x_str = number_to_string(self.pubkey.point.x(), order)
        if self.pubkey.point.y() & 1:
            return b"\x03" + x_str
        return b"\x02" + x_str

    def to_string(self, encoding = 'uncompressed'):
        if encoding == "compressed":
            return self._compressed_encode()
        return self._raw_encode()

    def to_pem(self):
        return der.topem(self.to_der(), "PUBLIC KEY")

    def to_der(self):
        order = self.pubkey.order
        x_str = number_to_string(self.pubkey.point.x(), order)
        y_str = number_to_string(self.pubkey.point.y(), order)
        point_str = b'\x00\x04' + x_str + y_str
        return der.encode_sequence(der.encode_sequence(encoded_oid_ecPublicKey,
                                                       self.curve.encoded_oid),
                                   der.encode_bitstring(point_str))

    def verify(self, signature, data, hashfunc=None, sigdecode=sigdecode_string):
        hashfunc = hashfunc or self.default_hashfunc
        digest = hashfunc(data).digest()
        return self.verify_digest(signature, digest, sigdecode)

    def verify_digest(self, signature, digest, sigdecode=sigdecode_string):
        if len(digest) > self.curve.baselen:
            raise BadDigestError("this curve (%s) is too short "
                                 "for your digest (%d)" % (self.curve.name,
                                                           8 * len(digest)))
        number = string_to_number(digest)
        r, s = sigdecode(signature, self.pubkey.order)
        sig = ecdsa.Signature(r, s)
        if self.pubkey.verifies(number, sig):
            return True
        raise BadSignatureError


class SigningKey:
    def __init__(self, _error__please_use_generate=None):
        if not _error__please_use_generate:
            raise TypeError("Please use SigningKey.generate() to construct me")

    @classmethod
    def generate(klass, curve=NIST192p, entropy=None, hashfunc=sha1):
        secexp = randrange(curve.order, entropy)
        return klass.from_secret_exponent(secexp, curve, hashfunc)

    # to create a signing key from a short (arbitrary-length) seed, convert
    # that seed into an integer with something like
    # secexp=util.randrange_from_seed__X(seed, curve.order), and then pass
    # that integer into SigningKey.from_secret_exponent(secexp, curve)

    @classmethod
    def from_secret_exponent(klass, secexp, curve=NIST192p, hashfunc=sha1):
        self = klass(_error__please_use_generate=True)
        self.curve = curve
        self.default_hashfunc = hashfunc
        self.baselen = curve.baselen
        n = curve.order
        assert 1 <= secexp < n
        pubkey_point = curve.generator * secexp
        pubkey = ecdsa.Public_key(curve.generator, pubkey_point)
        pubkey.order = n
        self.verifying_key = VerifyingKey.from_public_point(pubkey_point, curve,
                                                            hashfunc)
        self.privkey = ecdsa.Private_key(pubkey, secexp)
        self.privkey.order = n
        return self

    @classmethod
    def from_string(klass, string, curve=NIST192p, hashfunc=sha1):
        assert len(string) == curve.baselen, (len(string), curve.baselen)
        secexp = string_to_number(string)
        return klass.from_secret_exponent(secexp, curve, hashfunc)

    @classmethod
    def from_pem(klass, string, hashfunc=sha1):
        # the privkey pem file has two sections: "EC PARAMETERS" and "EC
        # PRIVATE KEY". The first is redundant.
        if isinstance(string, str):
            string = string.encode()
        privkey_pem = string[string.index(b'-----BEGIN EC PRIVATE KEY-----'):]
        return klass.from_der(der.unpem(privkey_pem), hashfunc)

    @classmethod
    def from_der(klass, string, hashfunc=sha1):
        # SEQ([int(1), octetstring(privkey),cont[0], oid(secp224r1),
        #      cont[1],bitstring])
        s, empty = der.remove_sequence(string)
        if empty != b'':
            raise der.UnexpectedDER("trailing junk after DER privkey: %s" %
                                    binascii.hexlify(empty))
        one, s = der.remove_integer(s)
        if one != 1:
            raise der.UnexpectedDER("expected '1' at start of DER privkey,"
                                    " got %d" % one)
        privkey_str, s = der.remove_octet_string(s)
        tag, curve_oid_str, s = der.remove_constructed(s)
        if tag != 0:
            raise der.UnexpectedDER("expected tag 0 in DER privkey,"
                                    " got %d" % tag)
        curve_oid, empty = der.remove_object(curve_oid_str)
        if empty != b'':
            raise der.UnexpectedDER("trailing junk after DER privkey "
                                    "curve_oid: %s" % binascii.hexlify(empty))
        curve = find_curve(curve_oid)

        # we don't actually care about the following fields
        #
        # tag, pubkey_bitstring, s = der.remove_constructed(s)
        # if tag != 1:
        #     raise der.UnexpectedDER("expected tag 1 in DER privkey, got %d"
        #                             % tag)
        # pubkey_str = der.remove_bitstring(pubkey_bitstring)
        # if empty != "":
        #     raise der.UnexpectedDER("trailing junk after DER privkey "
        #                             "pubkeystr: %s" % binascii.hexlify(empty))

        # our from_string method likes fixed-length privkey strings
        if len(privkey_str) < curve.baselen:
            privkey_str = b'\x00' * (curve.baselen - len(privkey_str)) + privkey_str
        return klass.from_string(privkey_str, curve, hashfunc)

    def to_string(self):
        secexp = self.privkey.secret_multiplier
        s = number_to_string(secexp, self.privkey.order)
        return s

    def to_pem(self):
        # TODO: "BEGIN ECPARAMETERS"
        return der.topem(self.to_der(), "EC PRIVATE KEY")

    def to_der(self):
        # SEQ([int(1), octetstring(privkey),cont[0], oid(secp224r1),
        #      cont[1],bitstring])
        encoded_vk = b'\x00\x04' + self.get_verifying_key().to_string()
        return der.encode_sequence(der.encode_integer(1),
                                   der.encode_octet_string(self.to_string()),
                                   der.encode_constructed(0, self.curve.encoded_oid),
                                   der.encode_constructed(1, der.encode_bitstring(encoded_vk)),
                                   )

    def get_verifying_key(self):
        return self.verifying_key

    def sign_deterministic(self, data, hashfunc=None,
                           sigencode=sigencode_string,
                           extra_entropy=b''):
        hashfunc = hashfunc or self.default_hashfunc
        digest = hashfunc(data).digest()

        return self.sign_digest_deterministic(
            digest, hashfunc=hashfunc, sigencode=sigencode,
            extra_entropy=extra_entropy)

    def sign_digest_deterministic(self, digest, hashfunc=None,
                                  sigencode=sigencode_string,
                                  extra_entropy=b''):
        """
        Calculates 'k' from data itself, removing the need for strong
        random generator and producing deterministic (reproducible) signatures.
        See RFC 6979 for more details.
        """
        secexp = self.privkey.secret_multiplier

        def simple_r_s(r, s, order, v):
            return r, s, order, v

        retry_gen = 0
        while True:
            k = rfc6979.generate_k(
                self.curve.generator.order(), secexp, hashfunc, digest,
                retry_gen=retry_gen, extra_entropy=extra_entropy)
            try:
                r, s, order, v = self.sign_digest(digest, sigencode=simple_r_s, k=k)
                break
            except RSZeroError:
                retry_gen += 1

        return sigencode(r, s, order, v)

    def sign(self, data, entropy=None, hashfunc=None, sigencode=sigencode_string, k=None):
        """
        hashfunc= should behave like hashlib.sha1 . The output length of the
        hash (in bytes) must not be longer than the length of the curve order
        (rounded up to the nearest byte), so using SHA256 with nist256p is
        ok, but SHA256 with nist192p is not. (In the 2**-96ish unlikely event
        of a hash output larger than the curve order, the hash will
        effectively be wrapped mod n).

        Use hashfunc=hashlib.sha1 to match openssl's -ecdsa-with-SHA1 mode,
        or hashfunc=hashlib.sha256 for openssl-1.0.0's -ecdsa-with-SHA256.
        """

        hashfunc = hashfunc or self.default_hashfunc
        h = hashfunc(data).digest()
        return self.sign_digest(h, entropy, sigencode, k)

    def sign_digest(self, digest, entropy=None, sigencode=sigencode_string, k=None):
        if len(digest) > self.curve.baselen:
            raise BadDigestError("this curve (%s) is too short "
                                 "for your digest (%d)" % (self.curve.name,
                                                           8 * len(digest)))
        number = string_to_number(digest)
        r, s, v = self.sign_number(number, entropy, k)
        return sigencode(r, s, self.privkey.order, v)

    def sign_number(self, number, entropy=None, k=None):
        # returns a pair of numbers
        order = self.privkey.order
        # privkey.sign() may raise RuntimeError in the amazingly unlikely
        # (2**-192) event that r=0 or s=0, because that would leak the key.
        # We could re-try with a different 'k', but we couldn't test that
        # code, so I choose to allow the signature to fail instead.

        # If k is set, it is used directly. In other cases
        # it is generated using entropy function
        if k is not None:
            _k = k
        else:
            _k = randrange(order, entropy)

        assert 1 <= _k < order
        sig = self.privkey.sign(number, _k)
        return sig.r, sig.s, sig.recovery_param



import functools
from typing import List, Optional, Sequence

from ... import keccak

from ..common import int_from_bytes
from ..constants import EC_ORDER
from ...starkware.crypto.signature import (
    ECSignature,
    private_to_stark_key,
    sign
    # verify
)
from ...starkware.crypto.fast_pedersen_hash import (
    pedersen_hash
)

MASK_250 = 2**250 - 1
HEX_PREFIX = "0x"


def _starknet_keccak(data: bytes) -> int:
    """
    A variant of eth-keccak that computes a value that fits in a Starknet field element.
    """
    return int_from_bytes(keccak.SHA3(data)) & MASK_250


# def pedersen_hash(left: int, right: int) -> int:
#     """
#     One of two hash functions (along with _starknet_keccak) used throughout Starknet.
#     """
#     return cpp_hash(left, right)


def compute_hash_on_elements(data: Sequence) -> int:
    """
    Computes a hash chain over the data, in the following order:
        h(h(h(h(0, data[0]), data[1]), ...), data[n-1]), n).

    The hash is initialized with 0 and ends with the data length appended.
    The length is appended in order to avoid collisions of the following kind:
    H([x,y,z]) = h(h(x,y),z) = H([w, z]) where w = h(x,y).
    """
    return functools.reduce(pedersen_hash, [*data, len(data)], 0)


def message_signature(
    msg_hash: int, priv_key: int, seed: Optional[int] = 32
) -> ECSignature:
    """
    Signs the message with private key.
    """
    return sign(msg_hash, priv_key, seed)


# def verify_message_signature(
#     msg_hash: int, signature: List[int], public_key: int
# ) -> bool:
#     """
#     Verifies ECDSA signature of a given message hash with a given public key.
#     Returns true if public_key signs the message.
#     """
#     sig_r, sig_s = signature
#     # sig_w = pow(sig_s, -1, EC_ORDER)
#     return verify(msg_hash=msg_hash, r=sig_r, s=sig_s, public_key=public_key)


def encode_uint(value: int, bytes_length: int = 32) -> bytes:
    return value.to_bytes(bytes_length, byteorder="big")


def encode_uint_list(data: List[int]) -> bytes:
    return b"".join(encode_uint(x) for x in data)


def get_bytes_length(value: int) -> int:
    return (value.bit_length() + 7) // 8



# Copyright (c) 2021 Emanuele Bellocchia
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""Module for WIF encoding/decoding."""

# Imports
from typing import Tuple, Union

from ..addr import P2PKHPubKeyModes
from ..base58 import Base58Decoder, Base58Encoder
from ..coin_conf import CoinsConf
from ..ecc import IPrivateKey, Secp256k1PrivateKey
from ..utils.misc import BytesUtils


# Alias for P2PKHPubKeyModes
WifPubKeyModes = P2PKHPubKeyModes


class WifConst:
    """Class container for WIF constants."""

    # Suffix to be added if the private key correspond to a compressed public key
    COMPR_PUB_KEY_SUFFIX: bytes = b"\x01"


class WifEncoder:
    """
    WIF encoder class.
    It provides methods for encoding to WIF format.
    """

    @staticmethod
    def Encode(priv_key: Union[bytes, IPrivateKey],
               net_ver: bytes = CoinsConf.BitcoinMainNet.ParamByKey("wif_net_ver"),
               pub_key_mode: WifPubKeyModes = WifPubKeyModes.COMPRESSED) -> str:
        """
        Encode key bytes into a WIF string.

        Args:
            priv_key (bytes or IPrivateKey)        : Private key bytes or object
            net_ver (bytes, optional)              : Net version (Bitcoin main net by default)
            pub_key_mode (WifPubKeyModes, optional): Specify if the private key corresponds to a compressed public key

        Returns:
            str: WIF encoded string

        Raises:
            TypeError: If pub_key_mode is not a WifPubKeyModes enum or
                       the private key is not a valid Secp256k1PrivateKey
            ValueError: If the key is not valid
        """
        if not isinstance(pub_key_mode, WifPubKeyModes):
            raise TypeError("Public key mode is not an enumerative of WifPubKeyModes")

        # Convert to private key to check if bytes are valid
        if isinstance(priv_key, bytes):
            priv_key = Secp256k1PrivateKey.FromBytes(priv_key)
        elif not isinstance(priv_key, Secp256k1PrivateKey):
            raise TypeError("A secp256k1 private key is required")

        priv_key = priv_key.Raw().ToBytes()

        # Add suffix if correspond to a compressed public key
        if pub_key_mode == WifPubKeyModes.COMPRESSED:
            priv_key += WifConst.COMPR_PUB_KEY_SUFFIX

        # Add net address version
        priv_key = net_ver + priv_key

        # Encode key
        return Base58Encoder.CheckEncode(priv_key)


class WifDecoder:
    """
    WIF encoder class.
    It provides methods for encoding to WIF format.
    """

    @staticmethod
    def Decode(wif_str: str,
               net_ver: bytes = CoinsConf.BitcoinMainNet.ParamByKey("wif_net_ver")) -> Tuple[bytes, WifPubKeyModes]:
        """
        Decode key bytes from a WIF string.

        Args:
            wif_str (str)            : WIF string
            net_ver (bytes, optional): Net version (Bitcoin main net by default)

        Returns:
            tuple[bytes, WifPubKeyModes]: Key bytes (index 0), public key mode (index 1)

        Raises:
            Base58ChecksumError: If the base58 checksum is not valid
            ValueError: If the resulting key is not valid
        """

        # Decode string
        priv_key_bytes = Base58Decoder.CheckDecode(wif_str)

        # Check net version
        if priv_key_bytes[0] != ord(net_ver):
            raise ValueError(
                f"Invalid net version (expected 0x{ord(net_ver):02X}, got 0x{priv_key_bytes[0]:02X})"
            )

        # Remove net version
        priv_key_bytes = priv_key_bytes[1:]

        # Remove suffix if correspond to a compressed public key
        if Secp256k1PrivateKey.IsValidBytes(priv_key_bytes[:-1]):
            # Check the compressed public key suffix
            if priv_key_bytes[-1] != ord(WifConst.COMPR_PUB_KEY_SUFFIX):
                raise ValueError(
                    f"Invalid compressed public key suffix (expected 0x{ord(WifConst.COMPR_PUB_KEY_SUFFIX):02X}, "
                    f"got 0x{priv_key_bytes[-1]:02X})"
                )
            # Remove it
            priv_key_bytes = priv_key_bytes[:-1]
            pub_key_mode = WifPubKeyModes.COMPRESSED
        else:
            if not Secp256k1PrivateKey.IsValidBytes(priv_key_bytes):
                raise ValueError(f"Invalid decoded key ({BytesUtils.ToHexString(priv_key_bytes)})")
            pub_key_mode = WifPubKeyModes.UNCOMPRESSED

        return priv_key_bytes, pub_key_mode



# Copyright (c) 2022 Emanuele Bellocchia
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""
Module for SLIP-0173 human-readable parts.
Not all the human-readable parts are defined, but only the used ones.
Reference: https://github.com/satoshilabs/slips/blob/master/slip-0173.md
"""


class Slip173:
    """
    SLIP-0173 class.
    It defines the human-readable parts in according to SLIP-0173.
    """

    AKASH_NETWORK: str = "akash"
    AXELAR: str = "axelar"
    BAND_PROTOCOL: str = "band"
    BINANCE_CHAIN: str = "bnb"
    BITCOIN_MAINNET: str = "bc"
    BITCOIN_REGTEST: str = "bcrt"
    BITCOIN_TESTNET: str = "tb"
    CERTIK: str = "certik"
    CHIHUAHUA: str = "chihuahua"
    CELESTIA: str = "celestia"
    COSMOS: str = "cosmos"
    DYDX: str = "dydx"
    ELROND: str = "erd"
    FETCH_AI: str = "fetch"
    HARMONY_ONE: str = "one"
    INJECTIVE: str = "inj"
    IRIS_NETWORK: str = "iaa"
    KAVA: str = "kava"
    LITECOIN_MAINNET: str = "ltc"
    LITECOIN_TESTNET: str = "tltc"
    NEUTRON: str = "neutron"
    OKEX_CHAIN: str = "ex"
    OSMOSIS: str = "osmo"
    SECRET_NETWORK: str = "secret"
    STAFI: str = "stafi"
    TERRA: str = "terra"
    ZILLIQA: str = "zil"



# Copyright (c) 2021 Emanuele Bellocchia
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""Module with some algorithm utility functions."""

# Imports
from bisect import bisect_left
from typing import Any, List, Union


class AlgoUtils:
    """Class container for algorithm utility functions."""

    @staticmethod
    def BinarySearch(arr: List,
                     elem: Any) -> int:
        """
        Binary search algorithm simply implemented by using the bisect library.

        Args:
            arr (list): list of elements
            elem (any): element to be searched

        Returns:
            int: First index of the element, -1 if not found
        """

        invalid_idx = -1

        i = bisect_left(arr, elem)
        if i != len(arr) and arr[i] == elem:
            return i

        return invalid_idx

    @staticmethod
    def Decode(data: Union[bytes, str],
               encoding: str = "utf-8") -> str:
        """
        Decode from bytes.

        Args:
            data (str or bytes): Data
            encoding (str)     : Encoding type

        Returns:
            str: String encoded to bytes

        Raises:
            TypeError: If the data is neither string nor bytes
        """
        if isinstance(data, str):
            return data
        if isinstance(data, bytes):
            return data.decode(encoding)
        raise TypeError("Invalid data type")

    @staticmethod
    def Encode(data: Union[bytes, str],
               encoding: str = "utf-8") -> bytes:
        """
        Encode to bytes.

        Args:
            data (str or bytes): Data
            encoding (str)     : Encoding type

        Returns:
            bytes: String encoded to bytes

        Raises:
            TypeError: If the data is neither string nor bytes
        """
        if isinstance(data, str):
            return data.encode(encoding)
        if isinstance(data, bytes):
            return data
        raise TypeError("Invalid data type")

    @staticmethod
    def IsStringMixed(data_str: str) -> bool:
        """
        Get if the specified string is in mixed case.

        Args:
            data_str (str): string

        Returns:
            bool: True if mixed case, false otherwise
        """
        return any(c.islower() for c in data_str) and any(c.isupper() for c in data_str)



# Copyright (c) 2021 Emanuele Bellocchia
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""Module for BIP84 coins configuration."""

# Imports
from bip_utils.addr import P2WPKHAddrEncoder
from bip_utils.bip.bip32 import Bip32KeyNetVersions, Bip32Slip10Secp256k1
from  import DER_PATH_NON_HARDENED_FULL, BipCoinConf
from bip_utils.coin_conf import CoinsConf
from bip_utils.slip.slip44 import Slip44


# Bitcoin key net version for main net (zpub / zprv)
_BIP84_BTC_KEY_NET_VER_MAIN: Bip32KeyNetVersions = Bip32KeyNetVersions(b"\x04\xb2\x47\x46",
                                                                       b"\x04\xb2\x43\x0c")
# Bitcoin key test net version for test net (vpub / vprv)
_BIP84_BTC_KEY_NET_VER_TEST: Bip32KeyNetVersions = Bip32KeyNetVersions(b"\x04\x5f\x1c\xf6",
                                                                       b"\x04\x5f\x18\xbc")


class Bip84Conf:
    """Class container for BIP84 configuration."""

    # Configuration for Bitcoin main net
    BitcoinMainNet: BipCoinConf = BipCoinConf(
        coin_names=CoinsConf.BitcoinMainNet.CoinNames(),
        coin_idx=Slip44.BITCOIN,
        is_testnet=False,
        def_path=DER_PATH_NON_HARDENED_FULL,
        key_net_ver=_BIP84_BTC_KEY_NET_VER_MAIN,
        wif_net_ver=CoinsConf.BitcoinMainNet.ParamByKey("wif_net_ver"),
        bip32_cls=Bip32Slip10Secp256k1,
        addr_cls=P2WPKHAddrEncoder,
        addr_params={
            "hrp": CoinsConf.BitcoinMainNet.ParamByKey("p2wpkh_hrp"),
        },
    )
    # Configuration for Bitcoin regtest
    BitcoinRegTest: BipCoinConf = BipCoinConf(
        coin_names=CoinsConf.BitcoinRegTest.CoinNames(),
        coin_idx=Slip44.TESTNET,
        is_testnet=True,
        def_path=DER_PATH_NON_HARDENED_FULL,
        key_net_ver=_BIP84_BTC_KEY_NET_VER_TEST,
        wif_net_ver=CoinsConf.BitcoinRegTest.ParamByKey("wif_net_ver"),
        bip32_cls=Bip32Slip10Secp256k1,
        addr_cls=P2WPKHAddrEncoder,
        addr_params={
            "hrp": CoinsConf.BitcoinRegTest.ParamByKey("p2wpkh_hrp"),
        },
    )
    # Configuration for Bitcoin test net
    BitcoinTestNet: BipCoinConf = BipCoinConf(
        coin_names=CoinsConf.BitcoinTestNet.CoinNames(),
        coin_idx=Slip44.TESTNET,
        is_testnet=True,
        def_path=DER_PATH_NON_HARDENED_FULL,
        key_net_ver=_BIP84_BTC_KEY_NET_VER_TEST,
        wif_net_ver=CoinsConf.BitcoinTestNet.ParamByKey("wif_net_ver"),
        bip32_cls=Bip32Slip10Secp256k1,
        addr_cls=P2WPKHAddrEncoder,
        addr_params={
            "hrp": CoinsConf.BitcoinTestNet.ParamByKey("p2wpkh_hrp"),
        },
    )

    # Configuration for Litecoin main net
    LitecoinMainNet: BipCoinConf = BipCoinConf(
        coin_names=CoinsConf.LitecoinMainNet.CoinNames(),
        coin_idx=Slip44.LITECOIN,
        is_testnet=False,
        def_path=DER_PATH_NON_HARDENED_FULL,
        key_net_ver=_BIP84_BTC_KEY_NET_VER_MAIN,
        wif_net_ver=CoinsConf.LitecoinMainNet.ParamByKey("wif_net_ver"),
        bip32_cls=Bip32Slip10Secp256k1,
        addr_cls=P2WPKHAddrEncoder,
        addr_params={
            "hrp": CoinsConf.LitecoinMainNet.ParamByKey("p2wpkh_hrp"),
        },
    )
    # Configuration for Litecoin test net
    LitecoinTestNet: BipCoinConf = BipCoinConf(
        coin_names=CoinsConf.LitecoinTestNet.CoinNames(),
        coin_idx=Slip44.TESTNET,
        is_testnet=True,
        def_path=DER_PATH_NON_HARDENED_FULL,
        key_net_ver=Bip32KeyNetVersions(b"\x04\x36\xf6\xe1",
                                        b"\x04\x36\xef\x7d"),   # ttub / ttpv
        wif_net_ver=CoinsConf.LitecoinTestNet.ParamByKey("wif_net_ver"),
        bip32_cls=Bip32Slip10Secp256k1,
        addr_cls=P2WPKHAddrEncoder,
        addr_params={
            "hrp": CoinsConf.LitecoinTestNet.ParamByKey("p2wpkh_hrp"),
        },
    )



from bip_utils.slip.slip32.slip32 import (
    Slip32DeserializedKey, Slip32KeyDeserializer, Slip32PrivateKeySerializer, Slip32PublicKeySerializer
)
from bip_utils.slip.slip32.slip32_key_net_ver import Slip32KeyNetVersions



# Copyright (c) 2021 Emanuele Bellocchia
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""Module with helper class for generic BIP coins configuration handling."""

# Imports
from typing import Any, Dict, Optional, Tuple, Type

from ...addr.iaddr_encoder import IAddrEncoder
from ...bip32 import Bip32Base, Bip32KeyNetVersions, Bip32PublicKey
from ...utils.conf import CoinNames as UtilsCoinNames


class BipCoinFctCallsConf:
    """Bip coin function calls configuration class."""

    m_fct_names: Tuple[str, ...]

    def __init__(self,
                 *args: str) -> None:
        """
        Construct class.

        Args:
            args (str): Function names to be called
        """
        self.m_fct_names = args

    def ResolveCalls(self,
                     pub_key: Bip32PublicKey) -> Any:
        """
        Resolve function calls and get the result.

        Args:
            pub_key (Bip32PublicKey object): Bip32PublicKey object

        Returns:
            Any: Result
        """
        res = pub_key
        for fct_call in self.m_fct_names:
            res = getattr(res, fct_call)()
        return res


class BipCoinConf:  # pylint: disable=too-many-instance-attributes
    """Bip coin configuration class."""

    m_coin_names: UtilsCoinNames
    m_coin_idx: int
    m_is_testnet: bool
    m_def_path: str
    m_key_net_ver: Bip32KeyNetVersions
    m_wif_net_ver: Optional[bytes]
    m_bip32_cls: Type[Bip32Base]
    m_addr_params: Dict[str, Any]
    # m_addr_cls: Type[IAddrEncoder]
    m_any_addr_params_fct_call: bool

    def __init__(self,  # pylint: disable=too-many-arguments
                 coin_names: UtilsCoinNames,
                 coin_idx: int,
                 is_testnet: bool,
                 def_path: str,
                 key_net_ver: Bip32KeyNetVersions,
                 wif_net_ver: Optional[bytes],
                 bip32_cls: Type[Bip32Base],
                #  addr_cls: Type[IAddrEncoder],
                 addr_params: Dict[str, Any]) -> None:
        """
        Construct class.

        Args:
            coin_names (CoinNames object)           : Coin names
            coin_idx (int)                          : Coin index
            is_testnet (bool)                       : Test net flag
            def_path (str)                          : Default path
            key_net_ver (Bip32KeyNetVersions object): Key net versions
            wif_net_ver (bytes)                     : WIF net version, None if not supported
            bip32_cls (Bip32Base class)             : Bip32 class
            addr_params (dict)                      : Address parameters
            addr_cls (IAddrEncoder class)           : Address class
        """
        self.m_coin_names = coin_names
        self.m_coin_idx = coin_idx
        self.m_is_testnet = is_testnet
        self.m_def_path = def_path
        self.m_key_net_ver = key_net_ver
        self.m_wif_net_ver = wif_net_ver
        self.m_bip32_cls = bip32_cls
        self.m_addr_params = addr_params
        self.m_any_addr_params_fct_call = any(
            (isinstance(param_val, BipCoinFctCallsConf) for param_val in addr_params.values())
        )
        # self.m_addr_cls = addr_cls

    def CoinNames(self) -> UtilsCoinNames:
        """
        Get coin names.

        Returns:
            CoinNames object: CoinNames object
        """
        return self.m_coin_names

    def CoinIndex(self) -> int:
        """
        Get coin index.

        Returns:
            int: Coin index
        """
        return self.m_coin_idx

    def IsTestNet(self) -> bool:
        """
        Get if test net.

        Returns:
            bool: True if test net, false otherwise
        """
        return self.m_is_testnet

    def DefaultPath(self) -> str:
        """
        Get the default derivation path.

        Returns:
            str: Default derivation path
        """
        return self.m_def_path

    def KeyNetVersions(self) -> Bip32KeyNetVersions:
        """
        Get key net versions.

        Returns:
            Bip32KeyNetVersions object: Bip32KeyNetVersions object
        """
        return self.m_key_net_ver

    def WifNetVersion(self) -> Optional[bytes]:
        """
        Get WIF net version.

        Returns:
            bytes: WIF net version bytes
            None: If WIF is not supported
        """
        return self.m_wif_net_ver

    def Bip32Class(self) -> Type[Bip32Base]:
        """
        Get the Bip32 class.

        Returns:
            Bip32Base class: Bip32Base class
        """
        return self.m_bip32_cls

    def AddrParams(self) -> Dict[str, Any]:
        """
        Get the address parameters.

        Returns:
            dict: Address parameters
        """
        return self.m_addr_params

    def AddrParamsWithResolvedCalls(self,
                                    pub_key: Bip32PublicKey) -> Dict[str, Any]:
        """
        Get the address parameters with resolved function calls.

        Args:
            pub_key (Bip32PublicKey object): Bip32PublicKey object

        Returns:
            dict: Address parameters
        """
        addr_params = self.AddrParams()

        # Just use the internal object if nothing to be resolved to avoid creating a new one
        if not self.m_any_addr_params_fct_call:
            return addr_params

        # Create a new dictionary with resolved function calls
        return {
            param_name: param_val.ResolveCalls(pub_key) if isinstance(param_val, BipCoinFctCallsConf) else param_val
            for param_name, param_val in addr_params.items()
        }

    # def AddrClass(self) -> Type[IAddrEncoder]:
    #     """
    #     Get the address class.

    #     Returns:
    #         IAddrEncoder class: Address class
    #     """
    #     return self.m_addr_cls



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: dydxprotocol/accountplus/params.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from v4_proto.gogoproto import gogo_pb2 as gogoproto_dot_gogo__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n%dydxprotocol/accountplus/params.proto\x12\x18\x64ydxprotocol.accountplus\x1a\x14gogoproto/gogo.proto\"M\n\x06Params\x12\x43\n\x17is_smart_account_active\x18\x01 \x01(\x08\x42\"\xf2\xde\x1f\x1eyaml:\"is_smart_account_active\"B?Z=github.com/dydxprotocol/v4-chain/protocol/x/accountplus/typesb\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'dydxprotocol.accountplus.params_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'Z=github.com/dydxprotocol/v4-chain/protocol/x/accountplus/types'
  _PARAMS.fields_by_name['is_smart_account_active']._options = None
  _PARAMS.fields_by_name['is_smart_account_active']._serialized_options = b'\362\336\037\036yaml:\"is_smart_account_active\"'
  _PARAMS._serialized_start=89
  _PARAMS._serialized_end=166
# @@protoc_insertion_point(module_scope)



import sys
import kucoin.ccxt as ccxt_module
sys.modules['ccxt'] = ccxt_module

from kucoin.ccxt import kucoin as KucoinSync
from kucoin.ccxt.async_support.kucoin import kucoin as KucoinAsync
from kucoin.ccxt.pro.kucoin import kucoin as KucoinWs
